# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Name:         alteration.py
# Purpose:      alteration class and methods
#
# Author:       Marco de Freitas
#
# Copyright:    Copyright © 2021 Marco de Freitas
#
# License:      GNU Lesser General Public License 3
# -----------------------------------------------------------------------------
from __future__ import annotations
from typing import Tuple

class Alteration:
    """This class represents note alterations.

    :param string: Note name
    :type string: string

    :var label: usual string representation: empty, single, multiple '#' or 'b'
    :var value: note offset in half steps  (#=1,##=2,b=-1,bb=-2 etc.).

    >>> sharp = Alteration("#")
    >>> print(sharp.value, sharp.label)
    1 #
    """


    def __init__(self, param):
        if param == None: # corner case, when value is the result of a function, sometimes result is None
            self.label = ''
            self.value = 0
        elif isinstance(param, int): # init with half-steps value +1 = #, -1 = b
            self.label, self.value = Alteration._from_integer(param)

        elif isinstance(param, str): # init from string short, AbsInterval or AbsNote init from string like "b3" or "A#"
            self.label, self.value = Alteration._from_string(param)
        else:
            raise ValueError

    @staticmethod
    def _regex_process_alteration(expr: str) -> int:
        """ !!expr start without the note name !!
        :param expr:
        :return:
        """
        plus_symbols = ['#', '+']
        minus_symbols = ['♭', 'b']
        plus = sum([expr.count(symbols) for symbols in plus_symbols])
        minus = sum([expr.count(symbols) for symbols in minus_symbols])
        return plus - minus

    @staticmethod
    def _label_from_int(integer):
        if integer > 0:
            label = integer * '#'
        elif integer < 0:
            label = abs(integer) * 'b'
        elif integer == 0:
            label = ""
        else:
            print("way out")
        return label

    @staticmethod
    def _from_string(string) ->Tuple:
        if string == '' or string == '♮':
            label = ''
            value = 0
        else:
            value = Alteration._regex_process_alteration(string)
            label = Alteration._label_from_int(value)
        return label, value

    def _from_integer( integer):
        return Alteration._label_from_int(integer), integer



    def __add__(self, other:Alteration):
        """
        >>> a=Alteration(1)
        >>> b=Alteration(-1)
        >>> c=a+b
        >>> print(c.value)
        0
        """
        return Alteration(self.value+other.value)



    def alter(self, label:str):
        """
        This function sets new alteration value.

        :param :label : New alteration label ("#","b" etc.)
        :type :label : string

        :Example:

        >>> double_flat=Alteration('bb')
        >>> print(double_flat.value, double_flat.label)
        -2 bb
        >>> double_flat.alter('#')
        >>> print(double_flat.value, double_flat.label)
        1 #
        """
        if label == 0 or None:
            self.label = ''
            self.value = 0
        elif isinstance(label, str):
            self.value = Alteration._regex_process_alteration(label)
        elif isinstance(label, int):
            self.value += label
        if self.value > 0:
            self.label = abs(self.value) * "#"
        elif self.value < 0:
            self.label = abs(self.value) * "b"
        else:
            self.label = ''


