# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Name:         chords.py
# Purpose:      chords class and methods
#
# Author:       Marco de Freitas
#
# Copyright:    Copyright © 2021 Marco de Freitas
#
# License:      GNU Lesser General Public License 3
# -----------------------------------------------------------------------------
"""
Musenot module for chords. AbsChords represent abstract chord structures. Basically, an Abschord is a container for
AbsIntervals where AbsIntervals represent constitutiv degrees as AbsInterval from root.
"""

from __future__ import annotations

from musenot.intervals import AbsInterval
from musenot.notes import AbsNote

CHORDS_STRUCT_TO_TYPE_NAME = {('1', '3', '5'): 'Major', ('1', 'b3', '5'): 'Minor', ('1', 'b3', 'b5'): 'Diminished',
                              ('1', '3', '#5'): 'Augmented'}

CHORDS_SHORT = {'Major': 'Maj', 'Minor': 'min', 'Diminished': 'dim', 'Augmented': "aug"}


def regex_from_short(string):
    """
    'C-7' converted in root and degrees list
    :param string: short notation
    :return:
    """
    # (([A-G]+)([b#]*)){1}(([-]?)|(Maj?))?([7]?)
    pass


class AbsChord:
    """
    AbsChord class represents chords in their abstract form wich is an list of intevals from the root.
    A major triad is a unisson, a major third and a perfect fifth. AbsChord instances store this relations as a list.

    >>> a=AbsChord(['1', 'b3', '5'])
    >>> print(a.degrees)
    [AbsInterval('1'), AbsInterval('b3'), AbsInterval('5')]
    >>> a.invert(0)
    [AbsInterval('1'), AbsInterval('b3'), AbsInterval('5')]
    >>> a.invert(1)
    [AbsInterval('b3'), AbsInterval('5'), AbsInterval('1')]
    >>> a.invert(2)
    [AbsInterval('5'), AbsInterval('1'), AbsInterval('b3')]
    >>> len(a)
    3

    AbsNOte alteration methods work on chord degrees
    >>> a[1]+=1
    >>> print(a.degrees)
    [AbsInterval('1'), AbsInterval('3'), AbsInterval('5')]
    >>> print(a.code)
    ['1', 4, '5']
    """

    def __init__(self, degrees_list):
        self.degrees = [AbsInterval(step) for step in degrees_list]
        self.code = degrees_list
        self.type = CHORDS_STRUCT_TO_TYPE_NAME[tuple(self.code)]  # major, minor, {('1','3','5'):'Major' etc...
        self.short = CHORDS_SHORT[self.type]  # C-, C7 etc

    # TODO init from string (label) 'm','M','7'

    def __len__(self):
        return len(self.code)

    def __getitem__(self, key):
        return self.degrees[key]

    def __setitem__(self, key: int, item: AbsInterval):
        if isinstance(item, AbsInterval):
            self.degrees[key] = item
            self.code[key] = len(item)
            # TODO root case item ==0
        else:
            raise TypeError('Accept only AbsIntervals as new item value')

    # Conceptual difference between chord definition (root, third, fifth etc), inversions (root position, firs inversion,
    # second inversion) and figuration (actual realisation with octaves, drops, spreads).
    # Actual version of invert() mutates AbsChord instance, witch doesn't make any sense at a conceptual level.
    # As a module method invet() would return inversions as an AbsInterval. This should be enough for user needs.
    # other option is to store actual inversion state as an instance attribute. Still no mutation.

    def invert(self, index: int):
        if index in range(len(self)):
            return self.degrees[index:] + self.degrees[:index]
        else:
            return ValueError('Inversion number must be lower than chord size')

    def __repr__(self):
        """
        >>> AbsChord(['1', 'b3', '5'])
        AbsChord(['1', 'b3', '5'])
        """
        return f"AbsChord({[item.short_hand for item in self.degrees]})"


class Chord(AbsChord):
    """
    >>> notes=[ 'c', 'e', 'g']
    >>> myChord = Chord(notes)
    >>> print(myChord.notes)
    [AbsNote('c'), AbsNote('e'), AbsNote('g')]
    >>> myChord.root
    AbsNote('c')
    >>> myChord.degrees
    [AbsInterval('1'), AbsInterval('3'), AbsInterval('5')]
    >>> len(myChord)
    3
    """

    def __init__(self, notes):
        self.notes = [AbsNote(item) for item in notes]
        degrees_list = [self.notes[0] - self.notes[i] for i in range(len(self.notes))]
        degrees_list = [item.short_hand for item in degrees_list]

        super().__init__(degrees_list)
        self.label = self.root.__str__().capitalize() + self.short

        self.code = [item.short_hand for item in self.degrees]

    @property
    def root(self):
        return self.notes[0]

    def __repr__(self):
        """
        >>> Chord(['c','e','g'])
        Chord(['c', 'e', 'g'])
        """
        chord_notes = [f"{note.__str__()}" for note in self.notes]
        return f"Chord({chord_notes})"

    def __str__(self):
        """
        >>> print(Chord(['c','e','g']))
        C Major
        """
        return f"{self.root.__str__().capitalize()} {self.type}"

    def __add__(self, other):
        """
        >>> a = Chord(['c','e','g']) + AbsInterval('b3')
        >>> print(a.notes)
        [AbsNote('eb'), AbsNote('g'), AbsNote('bb')]
        >>> a + AbsNote('d')
        ['eb', 'g', 'bb', 'd']
        """
        if isinstance(other, AbsInterval):
            notes = [(note + other) for note in self.notes]

            return Chord([note.__str__() for note in notes])
        elif isinstance(other, AbsNote):
            new = [item.__str__() for item in self.notes]
            new.append(other.__str__())
            return new


if __name__ == "__main__":
    import doctest

    doctest.testmod()
