# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Name:         intervals.py
# Purpose:      alteration class and methods
#
# Author:       Marco de Freitas
#
# Copyright:    Copyright © 2021 Marco de Freitas
#
## License:      GNU Lesser General Public License 3
# -----------------------------------------------------------------------------
"""
Musenot.intervals module provide abstractions for intervals in music theory context.
"""

# FIXME API design, started working with initialisation from melodic notation "b3" instead of proper interval short-hand
#  notation "m3". Both initialisation are implemented. Need to make decisions for the API. AbsInterval("m3") and some
#  method absinterval_from_melodic("b3").

from __future__ import annotations
from typing import Tuple
import re


from musenot.notes import  AbsNote
from musenot.alteration import Alteration

from musenot.default import INTERVALS, INTERVALS_types, MAJOR_DEGREES_semitones, ADJECTIVS_perf, ADJECTIVS_majmin, DEFAULT_SCALE_semitones



absterval_regex = re.compile('(-)?([#]+|[b]+)?([1-9]{1}[0-9]*)+(.*)')

# match.group(1) (-)? : descending or not
# match.group(2) ([#]+|[b]+)? : alterations, 0 or arbitrary long sequence
# match.group(3) ([1-9]{1}[0-9]*) : size
# match.group(4) (.*) : test

absterval_regex_explicit = re.compile('([mMAdP]+)(-)?([1-9]{1}[0-9]*)+(.*)')
# match.group(1) ([mMAd]+) : quality
# match.group(2) (-)? : descending
# match.group(3) ([1-9]{1}[0-9]*) : name
# match.group(4) (.*) : test


def short_notation_to_litteral(string) -> Tuple[bool, str, str, str ]:
    """

    :param string:
    :return:

    >>> short_notation_to_litteral('b3')
    (False, 'b', '3')
    >>> short_notation_to_litteral('-#4')
    (True, '#', '4')

    """
    match = absterval_regex.match(string)
    if match is None:
        raise ValueError('No argument')
    else:
        descending=bool(match.group(1))
        if match.group(2) is None:
            alteration = ''
        else:
            alteration=match.group(2)
        value_name=match.group(3)
        test=match.group(4)
    return descending, alteration, value_name


def explicit_to_litteral(string)-> Tuple [bool, str, str]:
    """
    other notation, as in music21, "M3","P5","d6", where interval quality is explicitly stated. Descending "P-5".

    :param string:
    :return:
    """
    match = absterval_regex_explicit.match(string)
    if match is None:
        raise ValueError('No argument')
    elif match.group(4):
        raise ValueError('Some unexpected character in string tail')
    else:
        quality=match.group(1)
        descending=bool(match.group(2))
        value_name=match.group(3)

    if INTERVALS_types[int(value_name)-1] == "Major":
        if quality == "M":
            alteration = 0
        elif quality == "m":
            alteration = -1
        elif quality == "A":
            alteration == 1
        elif quality == "d":
            alteration == -2
        else:
            raise ValueError("multiple dd or AA note handled by now")
    elif INTERVALS_types[int(value_name)-1] == "Perfect":
        if quality == "A":
            alteration = 1
        elif quality == "d":
            alteration = -1
        elif quality == "P":
            alteration = 0
        else:
            raise ValueError("multiple dd or AA note handled by now")
    else:
        raise ValueError("Unable to process quality string")
    return descending, alteration, value_name


class IntervalName():
    """
    This class represents IntervalNames. Can be initialized either by a interval number or an explicite name. See
    examples below.

    :param number: numeric representation of the interval captured by regex.
    :type number: String
    :param name: explicit name initialization.
    :type number: String

    :example:

    >>> IntervalName(1)
    unison

    >>> IntervalName(2)
    second

    >>> a=IntervalName(3)
    >>> print(a.index)
    2
    >>> print(a.type)
    Major

    >>> IntervalName('unison')
    unison

    >>> IntervalName('second')
    second

    >>> a=IntervalName('third')
    >>> print(a.index)
    2
    >>> print(a.label)
    third
    """




    def __init__(self, param):
        """

        :param param: either name or number
        :param type: str or int
        """
        if isinstance(param, int):
            self.index = param - 1

        elif isinstance(param, str):
            self.index = INTERVALS.index(param)
        else:
            raise TypeError('init param type should be int or str')

        self.label = INTERVALS[self.index]
        self.type = INTERVALS_types[self.index]
        self.value = MAJOR_DEGREES_semitones[self.index % 7] + (self.index // 7) * 12

    def __str__(self):
        return f"{self.label}"

    def __repr__(self):
        return f"{self.label}"

    def __add__(self, other):
        """
        Addition of :class: IntervalName

        :param other: IntervalName
        :return: IntervalName

        >>> second = IntervalName(2)
        >>> third = IntervalName(3)
        >>> second + third
        fifth
        >>> second + 1
        third
        """
        if isinstance(other, IntervalName):
            return IntervalName(self.index + other.index + 2)
        elif isinstance(other, int):
            return self + IntervalName(other)

    def __sub__(self, other):
        """
        Substraction of IntervalNames, i.e. intervalName from one to other

        :param other: IntervalName
        :return: IntervalName


        >>> second = IntervalName(2)
        >>> third = IntervalName(3)
        >>> third - second
        second

        Agnostic to direction (ascendind or descending)
        >>> second - third
        second
        """
        return IntervalName(abs(other.index - self.index)+1)



class AbsInterval():
    """
    Class for abstract intervals

    :param string: interval analytic description, for example '3' for major third
    :type strig: string



    >>> a=AbsInterval('5')
    >>> print(a)
    Perfect fifth

    >>> b=AbsInterval('b3')
    >>> print(b)
    minor third

    >>> print(a+b)
    minor seventh

    >>> print(AbsInterval('-3'))
    descending Major third

    >>> thirdmajor=AbsInterval('3')
    >>> thirdminor=AbsInterval('b3')
    >>> print(thirdmajor+thirdminor)
    Perfect fifth


    >>> c1=AbsInterval.from_description('third',4)
    >>> print(c1)
    Major third



    >>> c1=AbsInterval.from_description('third',4,True)
    >>> print(c1)
    descending Major third
    """



    def _init_from_args(self, args):
        if len(args)== 2 :
            self.descending, self.alteration_value, name = explicit_to_litteral(args[0])
            self.alteration_label=AbsInterval._alteration_label_from_int(self.alteration_value)
        else:
            self.descending, self.alteration_label , name = short_notation_to_litteral(args[0])
            self.alteration_value = AbsInterval._alteration_counter(self.alteration_label)
        self.name = IntervalName(int(name))
        self.half_steps = self.name.value + self.alteration_value
        if self.name.type == 'Perfect':
            self.quality = ADJECTIVS_perf[2 + self.alteration_value]
        elif self.name.type == 'Major':
            self.quality = ADJECTIVS_majmin[3 + self.alteration_value]

    def _init_from_kwargs(self, kwargs):
        try:
            self.descending = kwargs['descending']
        except KeyError:
            self.descending = False
        self.half_steps = abs(kwargs['size'])
        self.name = IntervalName(kwargs['name'])
        self.alteration_value = self.half_steps - self.name.value
        self.alteration_label = AbsInterval._alteration_label_from_int(self.alteration_value)
        if self.name.type == 'Perfect':
            self.quality = ADJECTIVS_perf[2 + self.alteration_value]
        else:
            self.quality = ADJECTIVS_majmin[3 + self.alteration_value]

    @staticmethod
    def _alteration_counter(string:str):
        if string is None:
            return 0
        else:
            plus = string.count('#')
            minus = string.count('b')
            value = plus-minus
            return value

    @staticmethod
    def _alteration_label_from_int(integer:int):
        if integer > 0:
            label = integer * '#'
        elif integer < 0:
            label = abs(integer) * 'b'
        elif integer == 0:
            label = ""
        return label

    def __init__(self, *args, **kwargs):
        """
        positional argument is melodic short notation, for example
        >>> minor_third = AbsInterval("b3")
        >>> descending_minor_third = AbsInterval("-b3")

        to use explicit interval notation add some second positional parameter, 'explicit' by convention, but any will work
        >>> major_third = AbsInterval("M3", 'explicit')
        >>> descending_major_third = AbsInterval("m-3", 'explicit')

        named arguments are size:int, name:str, descending:bool
        >>> print(AbsInterval(size=10, name='seventh', descending=False))
        minor seventh
        """
        if args:
            self._init_from_args(args)
        elif kwargs:
            self._init_from_kwargs(kwargs)
        else:
            raise ValueError('Cannot mix named parameters and string description')
        self.description = self._description()


    def _description(self) -> str:
        nb = self.name.index + 1
        if self.descending is True:
            return 'descending' + ' ' + self.quality + ' ' + self.name.label
        if self.descending is not True:
            return self.quality + ' ' + self.name.label

    @property
    def short_hand(self):
        return self._cifer()

    def _cifer(self) -> str:
        if self.descending is True:
            return self.alteration_label + str(self.name.index + 1)
        else:
            return self.alteration_label + str(self.name.index + 1)

    @staticmethod   # FIXME doct test errors origin AbsInterval from notes dont have Alteration label, probably due to AbsInterval.__init__('', args)
    def from_notes(qualified_name1:str, qualified_name2:str, descending=False, octave=0) -> AbsInterval:
        """
        >>> AbsInterval.from_notes('c','d')
        AbsInterval('2')
        >>> AbsInterval.from_notes('c','eb')
        AbsInterval('b3')
        >>> AbsInterval.from_notes('c','g',descending=True)
        AbsInterval('4')
        >>> AbsInterval.from_notes('c','g',octave=1)
        AbsInterval('12')
        >>> AbsInterval.from_notes('c','g',True,1)
        AbsInterval('11')
        """
        note1 = AbsNote(qualified_name1)
        note2 = AbsNote(qualified_name2)
        if descending:
            result = note2 - note1

        elif not descending:
            result = note1 - note2

        result.octave(octave)
        # if descending == False:
        #     name_index = (note2.name.index - note1.name.index  ) % 6
        #     size = (note2.indexInRefScale - note1.indexInRefScale ) % 11
        # if descending == True:
        #     name_index = (note1.name.index - note2.name.index + 7) %6
        #     size = (note1.indexInRefScale - note2.indexInRefScale -10) % 11
        # if octave != 0:
        #     size = size + (octave * 12)
        # name = INTERVALS[name_index + (octave * 7)]
        return result
        # return AbsInterval( size=size, name=name)

    @staticmethod
    def from_description(name, size, descending=False) -> AbsInterval:
        return AbsInterval(name=name, size=size, descending=descending)

    def octave(self, number):
        """

        :param number:
        :return:
        """
        if number == 0:
            pass
        else:
            self.half_steps += number * 12
            self.name= self.name + 7



    def __repr__(self):
        '''
        >>> AbsInterval('2')
        AbsInterval('2')
        '''
        if self.descending is True:
            return f"AbsInterval(\'-{self.alteration_label}{str(self.name.index + 1)}\')"
        else:
            return f"AbsInterval(\'{self.alteration_label}{str(self.name.index + 1)}\')"

    def __str__(self):

        if self.descending is True:
            if self.quality is None:
                return 'descending' + ' ' + self.name.label
            elif self.quality is not None:
                return 'descending' + ' ' + self.quality + ' ' + self.name.label
        else:
            if self.quality is None:
                return f"{self.name.label}"
            elif self.quality is not None:
                return f"{self.quality} {self.name.label}"

    # Rich comparaison methods

    def __lt__(self, other):
        """
        Comparison method, return True if self < other, mesured by actual size in half-steps

        >>> AbsInterval('2')<AbsInterval('3')
        True

        >>> AbsInterval('3')<AbsInterval('b4')
        False
        """
        if self.half_steps < other.half_steps:
            return True
        else:
            return False

    def __le__(self, other):
        """
        Comparison method, return True if self=<other

        >>> AbsInterval('2')<=AbsInterval('3')
        True

        >>> AbsInterval('3')<=AbsInterval('b4')
        True
        """
        if self.half_steps <= other.half_steps:
            return True
        else:
            return False

    def __eq__(self, other):
        """
        Return True if self==other

        >>> AbsInterval('#2')==AbsInterval('b3')
        True

        >>> AbsInterval('#4')==AbsInterval('b5')
        True

        >>> AbsInterval('3')!=AbsInterval('4')
        True
        """
        if self.half_steps == other.half_steps: # and self.descending == other.descending and self.name == other.name and self.quality == other.quality and self.index == other.index:
            return True
        else:
            return False

    def __gt__(self, other):
        """
        Return True if self>other

        >>> AbsInterval('3')>AbsInterval('2')
        True

        >>> AbsInterval('#4')>AbsInterval('b5')
        False

        """
        if self.half_steps > other.half_steps:
            return True
        else:
            return False

    def __ge__(self, other):
        """
        Return True if self>=other

        >>> AbsInterval('3')>=AbsInterval('2')
        True

        >>> AbsInterval('#4')>=AbsInterval('b5')
        True

        """
        if self.half_steps >= other.half_steps:
            return True
        else:
            return False

    # Emulating container types

    def __len__(self):
        """
        Returns the size of the AbsInterval mesured in half-steps.

        :returns: AbsInterval size in half-steps.
        :rtype: int

        >>> a=AbsInterval('#4')
        >>> print(len(a))





        """
        return self.half_steps

    # Emulating numeric types

    def add_absinterval(self, other : AbsInterval) -> AbsInterval:
        """

        :param other:
        :return:
        """

        first_length = (self.descending * -1 + (not self.descending) * 1) * self.half_steps
        first_name_index =(self.descending * -1 +  (not self.descending) * 1) * self.name.index
        second_length = (other.descending * -1 + (not other.descending) * 1 ) * other.half_steps
        second_name_index = (other.descending * -1 + (not other.descending) * 1) * other.name.index

        # if self.descending is False:
        #     first_length = self.half_steps
        #     first_name_index = self.name.index
        # if other.descending is False:
        #     second_length = other.half_steps
        #     second_name_index = other.name.index
        # if self.descending is True:
        #     first_length = -self.half_steps
        #     first_name_index = -self.name.index
        # if other.descending is True:
        #     second_length = -other.half_steps
        #     second_name_index = -other.name.index

        new_size = first_length + second_length
        new_name_index = abs(first_name_index + second_name_index)
        new_name = INTERVALS[new_name_index]
        if new_size < 0:
            descending = True
        else:
            descending = False
        # print(new_size, new_name_index)
        return AbsInterval(size=abs(new_size), name=new_name, descending=descending)


    def add_half_step(self, integer: int) -> AbsInterval:
        """

        :param other:
        :return:
        """
        size = self.half_steps + integer
        name = self.name.label
        return AbsInterval( size=size, name=name)

    def add_AbsNote(self, other: AbsNote) -> AbsNote:
        """

        :param other:
        :return:

        >>> AbsNote('c') + AbsInterval('-2')
        AbsNote('bb')
        >>> AbsNote('c') + AbsInterval('-b3')
        AbsNote('a')
        >>> AbsNote('a') + AbsInterval('-b3')
        AbsNote('f#')
        """
        if self.descending :
            new_indexInRefScale = (other.indexInRefScale  - self.half_steps) % 12
            new_name_index = (other.name.index - self.name.index ) % 7
        elif not self.descending:
            new_indexInRefScale = (other.indexInRefScale + self.half_steps) % 12
            new_name_index = (self.name.index + other.name.index) % 7

        if new_name_index == 0 and new_indexInRefScale != 0:
            unaltered_index_in_ref_scale = 12
        else:
            unaltered_index_in_ref_scale = DEFAULT_SCALE_semitones[new_name_index]
        # print( new_indexInRefScale, unaltered_index_in_ref_scale )
        delta_index = new_indexInRefScale - unaltered_index_in_ref_scale
        # print(delta_index)
        return AbsNote("", new_name_index, delta_index)



    def __add__(self, other):
        """Return summe of two intervals

        :param other: AbsInterval at right place of the summation
        :typye other: AbsInterval


        Summ of two AbsIntervals is a AbsInterval.
        >>> AbsInterval('3') + AbsInterval('b3')
        AbsInterval('5')

        Summation works as the usual compostion of intervals. Regarding addition, unison is the neutral element.
        >>> AbsInterval('1') + AbsInterval('2')
        AbsInterval('2')

        Summation does no assumption about enharmonic conversion. This means that AbsInterval codes must be carrefuly
        selected. In the following example both results have same size in half-steps but not the same conceptual
        meaning in term of degrees. AbsInterval('#4') is a raised fourth, AbsInterval('b5') is a lowered fifth. >>>
        AbsInterval('b3') + AbsInterval('#2') AbsInterval('#4')

        >>> AbsInterval('b3') + AbsInterval('b3')
        AbsInterval('b5')

        >>> print(AbsInterval('b2') + AbsInterval('#2'))
        Major third

        >>> print(AbsInterval('#4') + AbsInterval('b4'))
        minor seventh

        >>> print(AbsInterval('-2') + AbsInterval('-5'))
        descending Major sixth


        Summ of AbsInterval and AbsNote returns a AbsNote.
        >>> AbsInterval('b3') + AbsNote('c')
        AbsNote('eb')

        >>> AbsInterval('1')+AbsNote('c')
        AbsNote('c')

        >>> AbsNote('c')+ AbsInterval('1')
        AbsNote('c')

        >>> AbsInterval('1')+AbsNote('e')
        AbsNote('e')

        >>> AbsInterval('1')+AbsNote('eb')
        AbsNote('eb')

        >>> AbsNote('g') + AbsInterval('3')
        AbsNote('b')

        >>> AbsNote('f')+ AbsInterval('b3')
        AbsNote('ab')

        >>> AbsInterval('bb3')+AbsNote('f')
        AbsNote('abb')


        >>> AbsInterval('b3')+1
        AbsInterval('3')
        """

        if isinstance(other, AbsInterval):
             return  self.add_absinterval(other)
        elif isinstance(other, int):
            return self.add_half_step(other)
        elif isinstance(other, AbsNote):
            return self.add_AbsNote(other)
        else:
            return NotImplemented

    def __radd__(self, left):
        """Add some half-steps to an AbsInterval

        :param left: Nb of half-steps to be added
        :type left: int

        >>> a=AbsInterval('b3')+1
        >>> print(a)
        Major third

        >>> b=AbsInterval('b3')
        >>> 1 + b
        AbsInterval('3')




        >>> AbsNote('c')+ AbsInterval('5')
        AbsNote('g')
        >>> AbsNote('db')+ AbsInterval('3')
        AbsNote('f')

        """

        if isinstance(left, int):
            return self.__add__(left)

        if isinstance(left, AbsNote):
            return self.__add__(left)

        else:
            return NotImplemented

# FIXME substraction expected to return also descneding intervals, sub is used in AbsInterval.invert()
    def __sub__(self, other):
        """Substraction

        >>> AbsInterval('4')-AbsInterval('2')
        AbsInterval('b3')

        >>> AbsInterval('2')-AbsInterval('4')
        AbsInterval('-b3')

        >>> print(AbsInterval('2')-AbsInterval('4'))
        descending minor third



        """
        if isinstance(other, AbsInterval):
            size = self.half_steps - other.half_steps
            new_name_index = self.name.index - other.name.index
            if size >= 0:
                new_name = INTERVALS[new_name_index]
                return AbsInterval(size=size, name=new_name)
            if size < 0:
                new_name = INTERVALS[abs(new_name_index)]
                size = abs(size)
                return AbsInterval( size=size, name=new_name, descending=True)
        else:
            return NotImplemented

    def __len__(self):
        """
        >>> z=AbsInterval('3')
        >>> len(z)
        4
        """

        return self.half_steps

    def __invert__(self):
        """
        >>> a=AbsInterval('-3')
        >>> a = ~a
        >>> print(a.short_hand)
        3
        """
        self.descending = not self.descending
        return self

    def invert(self):
        """
        >>> a=AbsInterval('3')
        >>> a.invert()
        AbsInterval('b6')
        >>> print(a.invert())
        minor sixth


        >>> a+a.invert()
        AbsInterval('8')

        """
        return AbsInterval('8') - self


# if __name__ == "__main__":
#     import doctest
#
#     doctest.testmod()
