# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Name:         notename.py
# Purpose:      note name class and methods
#
# Author:       Marco de Freitas
#
# Copyright:    Copyright © 2021 Marco de Freitas
#
# License:      GNU Lesser General Public License 3
# -----------------------------------------------------------------------------
"""
This module provides note name class and methods. NoteName object are used thru  composition by nearly any other class
of musenot library.

In english we use the seven first letters [a..g], other languages use other naming convention. Creating a dedicated
module should ease traduction work.

NoteName objects are initialized with a explicite string type parametre "a" or index value (considering note position in
A minor natural scale).

Note names cycle in a upward or downward direction.
NoteName.diff() offer counting capabilities in both directions. This basic arithmetic is relevant for interval naming
and at the root all music theory knowledge.
"""
from __future__ import annotations

from musenot.default import NOTE_NAMES


# Place this constant value in DEFAULT? Argument against, those are specific to this module

NOTE_NAMES_WHEEL = dict(a=["a", "b", "c", "d", "e", "f", "g"], b=["b", "c", "d", "e", "f", "g", "a"],
                       c=["c", "d", "e", "f", "g", "a", "b"], d=["d", "e", "f", "g", "a", "b", "c"],
                       e=["e", "f", "g", "a", "b", "c", "d"], f=["f", "g", "a", "b", "c", "d", "e"],
                       g=["g", "a", "b", "c", "d", "e", 'f'])

NOTE_NAMES_WHEEL_DESCENDING = dict(a=["a", "g", "f", "e", "d", "b",  "c"], b=["b", "a", "g", "f", "e", "d", "c"],
                       c=["c", "b", "a", "g", "f", "e", "d"], d=["d", "c", "b", "a", "a", "f", "e"],
                       e=["e", "d", "c", "d", "a", "g", "f"], f=["f", "e", "d", "c", "b", "a", "g"],
                       g=["g", "f", "e", "d", "c", "b", 'a'])

REF_SCALE = dict(a=0, b=2, c=3, d=5, e=7, f=8, g=10)


class NoteName():
    """
    >>> a_note_name = NoteName('c')
    >>> a_note_name
    NoteName('c')
    >>> print(a_note_name)
    c
    >>> print(a_note_name.index)
    2
    >>> print(a_note_name.index_in_ref_scale)
    3
    >>> an_other_note_name = NoteName(0)
    >>> an_other_note_name
    NoteName('a')
    >>> print(an_other_note_name.index)
    0
    >>> print(an_other_note_name.index_in_ref_scale)
    0

    """
    def __init__(self, param):
        if isinstance(param, str):
            self.label, self.index, self.index_in_ref_scale = NoteName._from_name(param)
        elif isinstance(param, int):
            self.label, self.index, self.index_in_ref_scale = NoteName._from_index(param)

    @staticmethod
    def _from_name(name: str):
        if name not in NOTE_NAMES:
            raise ValueError(f"Allowed Note are {NOTE_NAMES}")
        else:
            label = name
            index = NOTE_NAMES.index(label)
            index_in_ref_scale = REF_SCALE[label]
        return label, index, index_in_ref_scale

    @staticmethod
    def _from_index(index: int):
        if not 7 >= index >= 0:
            raise ValueError("Allowed indexes are 0..7")
        else:
            index = index
            label = NOTE_NAMES[index]
            index_in_ref_scale = REF_SCALE[label]
        return label, index, index_in_ref_scale

    def __str__(self):
        return f"{self.label}"


    def __repr__(self):
        return f"{self.__class__.__name__}('{self.label}')"


    def diff(self, other, descending=None):
        """
        distance in note names cycle, ascending default.

        :param other: Other NoteName
        :other type:class:NoteName
        :param descending: Descending cycle
        :return: int

        >>> c_note_name = NoteName('c')
        >>> d_note_name = NoteName('d')
        >>> c_note_name.diff(d_note_name)
        1
        >>> d_note_name.diff(c_note_name)
        6
        >>> c_note_name.diff(d_note_name, descending=True)
        -6
        >>> d_note_name.diff(c_note_name, descending=True)
        -1
        """

        if isinstance(other, NoteName):
            if descending is None:
                 return NOTE_NAMES_WHEEL[self.label].index(other.label)
            else:
                 return - NOTE_NAMES_WHEEL_DESCENDING[self.label].index(other.label)


    def __sub__(self, other):
        if isinstance(other, int):
            self.index -= other
            self.label = NOTE_NAMES[self.index]
            self.index_in_ref_scale = REF_SCALE[self.label]

    def __add__(self, other):
        if isinstance(other, int):
            self.index += other
            self.label = NOTE_NAMES[self.index]
            self.index_in_ref_scale = REF_SCALE[self.label]

