# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Name:         notes.py
# Purpose:      classes for notes in music theory (determined or
#               undetermined octave)
#
# Author:       Marco de Freitas
#
# Copyright:    Copyright © 2021 Marco de Freitas
#
# License:      GNU Lesser General Public License 3
# -----------------------------------------------------------------------------

"""Provides classes for notes in music theory. AbsNote classe for notes with undetermined octave. Note classe for notes
with determined octave. """

from __future__ import annotations

from musenot.default import INTERVALS
from musenot.alteration import Alteration
from musenot.notename import NoteName

class AbsNote():
    """
    Class represents abstract notes used in music theory to discuss
    general relations.


    :var name: Note nane properties and methods
    :type name: NoteName object
    :var alteration: Alteration properties and methods
    :type alteration: Alteration Object
    """

    def __init__(self, abs_note_name, index=None, alteration=None):
        """
                :param abs_note_name: Qualified note name like 'c', 'd#', 'eb'
                :type abs_note_name: String


                >>> x = AbsNote('c')
                >>> x.name.label
                'c'
                >>> x.indexInRefScale
                3
                """
        if index == None and alteration == None:

            self.name = NoteName(abs_note_name[0])
            self.alteration = Alteration(abs_note_name[1:])
        else:
            self.name = NoteName(index)
            self.alteration = Alteration(alteration)


    @property
    def indexInRefScale(self) -> int:
        """
        >>> AbsNote('ab').indexInRefScale
        11
        """
        return (self.name.index_in_ref_scale + self.alteration.value) % 12

    def __str__(self):
        """
        >>> a =AbsNote('c')
        >>> print(a)
        c
        """
        return f"{self.name.label}{self.alteration.label}"


    def __repr__(self):
        """
        >>> AbsNote('c')
        AbsNote('c')
        """
        return f"AbsNote('{self.__str__()}')"

    def __eq__(self, other):
        if self.indexInRefScale == other.indexInRefScale:
            return True

    def sub_note(self, other:AbsNote) ->AbsInterval:
        """
        AbsNote substraction. Used by AbsNote.__sub__ after isntance type checking. Use it at your own risks.

        >>> a = AbsNote('c')
        >>> b = AbsNote('d')
        >>> a-b
        AbsInterval('2')
        >>> b-a
        AbsInterval('b7')
        >>> AbsNote('g') - AbsNote('f')
        AbsInterval('b7')
        >>> AbsNote('a') -AbsNote('a')
        AbsInterval('1')

        >>> AbsNote('g') - AbsNote('c')
        AbsInterval('4')
        """
        from musenot.intervals import AbsInterval

        size = ( other.indexInRefScale - self.indexInRefScale ) % 12

        name = INTERVALS[self.name.diff(other.name)]
        return AbsInterval( name=name, size=size, descending=False)


    def is_enharmonic(self, other:AbsNote) -> bool:
        """
            Test if AbsNote instance is enharmonic with other note represented in short hand notation or as Absterval
            instance.

            >>> this_note = AbsNote('b')
            >>> that_note = AbsNote('cb')
            >>> this_note.is_enharmonic(that_note)
            True
            >>> this_other_note = AbsNote('d')
            >>> that_note.is_enharmonic(this_other_note)
            False
        """

        return (self.indexInRefScale  == other.indexInRefScale)



    def sub_interval(self, interval:AbsInterval) -> AbsInterval:
        from musenot.intervals import AbsInterval
        return interval.__sub__(self)


    def __sub__(self, other):
        """
        See documentation of sub_notes
        """
        if isinstance(other, AbsNote):
            return self.sub_note(other)
        elif isinstance(other, AbsInterval):
            return self.sub_interval(other)



    def alter(self, alteration):
        """
        Either incremental from int (+1, -1 etc.) or set new value from string ("b", "##")
        :param alteration:
        :return:
        """
        if isinstance(alteration, int):
            self.alteration += alteration
        elif isinstance(alteration, int):
            self.alteration.alter(alteration)


class Note(AbsNote):
    """
    >>> a=Note('a#',2)
    >>> print(a.octave)
    2
    """


    def __init__(self, string, octave):
        super().__init__(string)
        self.octave = octave



####    def __str__(self):
####        note=super().__str__()
####        octave=self.octave
####        return  note + octave 
##
##
##    def __sub__(self,other):
##        '''
##        >>> Note('a',4)-Note('g#',4)
##        Absterval('b2')
##        
##        >>> Note('c',3)-Note('a',3)
##        Absterval('b3')
##
##        
##        '''
####        >>> Note('g',3) - Note('g',4)
##        
##        from musenot.intervals import Absterval
##        delta_octave= self.octave - other.octave
##        if isinstance(other,Note):
####            if descending == False:
####                name_index = (note2.indexInNamesWheel-note1.indexInNamesWheel) % 7
####                size = (note2.indexInRefScale-note1.indexInRefScale) % 12
####            else:
####                name_index = (note1.indexInNamesWheel-note2.indexInNamesWheel) % 7
####                size = (note1.indexInRefScale-note2.indexInRefScale) % 12
####            if octave != 0:
####                size = size+octave*12
####            name = INTERVALS[name_index+ octave*7] 
####            return Absterval('', name=name, size=size)
##
##            
##
##            if delta_octave >= 0:# name.index> other etc.!!!!!!!
##                name_index = (self.name.index - other.name.index) % 7 + delta_octave*7
##            else:
##                name_index = (other.name.index - self.name.index) % 7 + delta_octave*7
##                descending=True
##            name = INTERVALS[name_index]
##            size = (self.name.indexInRefScale + self.alteration.value - other.name.indexInRefScale
##                    - other.alteration.value) % 12
##            return Absterval('',name=name, size=size)
##        else:
##            return NotImplemented
##        
### To do class Note:
### attribut: name, alteration, octave
### example: C#4
### ou une seule classe avec méthode pour ajouter octave?


if __name__ == "__main__":
    import doctest
    doctest.testmod()
