# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Name:         scales.py
# Purpose:      scales class and methods
#
# Author:       Marco de Freitas
#
# Copyright:    Copyright © 2021 Marco de Freitas
#
# License:      GNU Lesser General Public License 3
# -----------------------------------------------------------------------------
from __future__ import annotations

from typing import List

from musenot.intervals import AbsInterval
from musenot.default import INTERVALS
from musenot.notes import AbsNote
from musenot.chords import AbsChord, Chord


class Abscale:
    """
    This is a class representation of abstract scales, i.e a structure of intervals with no determined note names.

    :var code: Successiv scale intervals length mesured in half-steps, string type.
    :var steps: Corresponding list of seconds intervals, represented by a list of :class:musenot.intervals.AbsInterval.
    :var degrees: Step interval from root, represented by a list of :class:`musenot.intervals.AbsInterval`.



    Creation of an abstract scale is done by providing scale steps measured in half-steps
    >>> major=Abscale('2212221')
    >>> minor=Abscale('2122122')

    Scale degrees is a scale representation where each note is represented as an interval from the root
    >>> print(major.degrees)
    [AbsInterval('1'), AbsInterval('2'), AbsInterval('3'), AbsInterval('4'), AbsInterval('5'), AbsInterval('6'), \
AbsInterval('7'), AbsInterval('8')]

    >>> print(minor.degrees)
    [AbsInterval('1'), AbsInterval('2'), AbsInterval('b3'), AbsInterval('4'), AbsInterval('5'), AbsInterval('b6'), \
AbsInterval('b7'), AbsInterval('8')]

    Scale structure can be represented as a list of successive intervals called steps
    >>> print(major.steps)
    [AbsInterval('2'), AbsInterval('2'), AbsInterval('b2'), AbsInterval('2'), AbsInterval('2'), AbsInterval('2'), \
AbsInterval('b2')]

    >>> print(minor.steps)
    [AbsInterval('2'), AbsInterval('b2'), AbsInterval('2'), AbsInterval('2'), AbsInterval('b2'), AbsInterval('2'), \
AbsInterval('2')]

    Scale steps structure can be divided in two tetracords with an intermediate interval
    >>> print(major.tetracord1)
    ['2', '2', 'b2']
    >>> print(major.tetracord2)
    ['2', '2', 'b2']
    >>> print(major.inter_tetracord)
    ['2']

    Scale steps and scale degrees are AbsInterval objects, AbsInterval methods can be called on Abscale.steps and \
Abscale.degrees
    >>> major.steps[0]+major.steps[1]+major.steps[2]
    AbsInterval('4')
    >>> print(major.steps[0]+major.steps[1]+major.steps[2])
    Perfect fourth
    >>> major.degrees[1]-major.degrees[0]
    AbsInterval('2')
    >>> print(major.degrees[1]-major.degrees[0])
    Major second
    """

# TODO handle non-heptatonic scales
    def __init__(self, code):
        self.verify_total(code)
        self.code = code
        int_code = [int(i) for i in code]
        self.steps = [AbsInterval(name=INTERVALS[1], size=i) for i in int_code]
        self.degrees = [AbsInterval(name=INTERVALS[i], size=sum(int_code[:i])) for i in range(len(int_code) + 1)]

    @property
    def tetracord1(self):
        if len(self) == 7:
            return [item.short_hand for item in self.steps[0:3]]
        else:
            raise NotImplementedError('tetracords are relevant only with heptatonic scales')

    @property
    def tetracord2(self):
        if len(self) == 7:
            return [item.short_hand for item in self.steps[4:8]]
        else:
            raise NotImplementedError('tetracords are relevant only with heptatonic scales')

    @property
    def inter_tetracord(self):
        if len(self) == 7:
            return [item.short_hand for item in self.steps[3:4]]
        else:
            raise NotImplementedError('tetracords are relevant only with heptatonic scales')

    @staticmethod
    def verify_total(code: str) -> bool:
        """Check if semitones summ is 12
        arguments:
        :param code: Size of the intervals of the scale
        :type code: String

        >>> Abscale.verify_total('2212221')
        True
        >>> Abscale.verify_total('221222')
        Traceback (most recent call last):
         ...
        ValueError: Total of semitones must be 12, unable to create the scale
        """

        if sum(int(i) for i in code) != 12:
            raise ValueError("Total of semitones must be 12, unable to create the scale")
        else:
            return True

    def __repr__(self):
        """
        Representation of Abscale instance in shell mode

        >>> major=Abscale('2212221')
        >>> minor=Abscale('2122122')
        >>> major
        Abscale('2212221')
        >>> minor
        Abscale('2122122')
        """
        return f"Abscale('{self.code}')"

    def __str__(self):
        """
        Print out scale degrees

        >>> major=Abscale('2212221')
        >>> minor=Abscale('2122122')

        >>> print(major)
        1, 2, 3, 4, 5, 6, 7, 8
        >>> print(minor)
        1, 2, b3, 4, 5, b6, b7, 8


        """
        liste = [item.short_hand for item in self.degrees]

        return f"{', '.join(liste)}"

    def __getitem__(self, key):
        """
        >>> major=Abscale('2212221')
        >>> major[0]
        AbsInterval('2')

        """
        return self.steps[key]

    def __setitem__(self, key, item):

        """
        >>> scale=Abscale('2212221')
        >>> print(scale)
        1, 2, 3, 4, 5, 6, 7, 8
        >>> scale[2]=AbsInterval('b3')
        >>> print(scale)
        1, 2, b3, 4, 5, 6, 7, 8
        >>> scale[3]=AbsInterval('#4')
        >>> print(scale)
        1, 2, b3, #4, 5, 6, 7, 8
        """
        if key == 0:
            raise ValueError
        else:

            self.degrees[key] = item
            self.steps[key] = item - self.steps[key - 1]
            self.code = self.code[:key] + str(len(self.steps[key])) + self.code[:key]

    def __contains__(self, other):
        """
        Implements membership test operators (like x in y), test if an AbsInterval is in scale.degrees

        >>> major=Abscale('2212221')
        >>> AbsInterval('3') in major
        True
        >>> AbsInterval('#5') in major
        False
        """
        if other in self.degrees:
            return True
        else:
            return False

    def __len__(self):
        return len(self.code)

    def harmonization(self, degree:int, size=3) -> AbsChord:
        """
        >>> major=Abscale('2212221')
        >>> chord=major.harmonization(1,3)
        >>> print(chord.degrees)
        [AbsInterval('1'), AbsInterval('3'), AbsInterval('5')]
        >>> print(chord.code)
        ['1', '3', '5']
        """

        degree = degree - 1
        infrastructure = [self.degrees[i + degree].short_hand for i in range(len(self.degrees)) if i % 2 == 0]
        superstructure = [self.degrees[i + degree].short_hand for i in range(len(self.degrees)) if i % 2 != 0]
        harmonization = infrastructure + superstructure
        return AbsChord(harmonization[0:size])

    def alter(self, key, value):
        """
        >>> major=Abscale('2212131')
        >>> major.alter(2,-1)
        >>> print(major)
        1, 2, b3, 4, 5, b6, 7, 8
        """
        if key == 0:
            return

        else:
            code_value = int(self.code[key - 1])

            new_value = code_value + value

            if new_value != 0:
                self.code = self.code[:key - 1] + str(code_value + value) + self.code[key + 1:]
                self.degrees[key] += value
        return


class Scale(Abscale):
    """
    >>> a=Scale('2212221', 'c')
    >>> print(a.root)
    c


    """

    def __init__(self, code, root):
        super().__init__(code)
        self.root = AbsNote(root)

        self.notes = [self.root]
        for i in range(len(code)):
            self.notes.append(self.notes[i] + self.steps[i])

    def __repr__(self):
        """
        Representation of Abscale instance in shell mode

        >>> major=Scale('2212221', 'c')
        >>> minor=Scale('2122122', 'a')
        >>> major
        Scale('2212221','c')
        >>> minor
        Scale('2122122','a')
        """
        return f"Scale('{self.code}','{self.notes[0]}')"

    def __str__(self):
        """
        >>> print(Scale('2212221','c'))
        c,d,e,f,g,a,b,c
        """
        result = [self.notes[i].__str__() for i in range(len(self.notes))]
        return f"{','.join(result)}"

    def __contains__(self, other):
        """
        >>> scale=Scale('2212221', 'c')
        >>> note=AbsNote('d')
        >>> note in scale
        True
        """

        def myfunct(x):
            if x == other:
                return True

        return filter(myfunct, self)

    def __getitem__(self, index):
        """
        >>> a=Scale('2212221', 'c')
        >>> a.notes[0]
        AbsNote('c')
        >>> a[0]
        AbsNote('c')
        """
        return self.notes[index]

    def transpose(self, interval):
        """
        >>> a=Scale('2212221', 'c')
        >>> a.transpose('b3')
        >>> print(a)
        eb,f,g,ab,bb,c,d,eb
        >>> a.transpose('-b3')
        >>> print(a)
        c,d,e,f,g,a,b,c
        >>> a.transpose('3')
        >>> print(a)
        e,f#,g#,a,b,c#,d#,e
        """
        transposition = AbsInterval(interval)
        self.notes = [note + transposition for note in self.notes]

    def harmonization(self, degree, size=3):
        """
        >>> a=Scale('2212221', 'c')
        >>> b=a.harmonization(1,3)
        >>> print(b.notes)
        [AbsNote('c'), AbsNote('e'), AbsNote('g')]


        """
        degree = degree - 1
        infrastructure = [self.notes[i + degree].__str__() for i in range(len(self.notes)) if i % 2 == 0]
        superstructure = [self.notes[i + degree].__str__() for i in range(len(self.notes)) if i % 2 != 0]
        harmonization = infrastructure + superstructure
        return Chord(harmonization[0:size])

    def __setitem__(self, index, value):
        pass


def scaleify(notes):
    """
    >>> a=scaleify(('a','b','c','d','e','f','g','a'))
    >>> print(a.notes)
    [AbsNote('a'), AbsNote('b'), AbsNote('c'), AbsNote('d'), AbsNote('e'), AbsNote('f'), AbsNote('g'), AbsNote('a')]
    >>> print(a.degrees)
    [AbsInterval('1'), AbsInterval('2'), AbsInterval('b3'), AbsInterval('4'), AbsInterval('5'), AbsInterval('b6'), \
AbsInterval('b7'), AbsInterval('8')]

    """
    absterval_list = [AbsNote(notes[i-1]) - AbsNote(notes[i]) for i in range(1, len(notes))]
    code = ''.join([str(item.__len__()) for item in absterval_list])
    return Scale(code, 'a')


if __name__ == "__main__":
    import doctest

    doctest.testmod()
